package eu.trade.client.normalizer;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.springframework.jdbc.core.ResultSetExtractor;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.datasource.DataSourceTransactionManager;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.TransactionException;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.support.DefaultTransactionDefinition;

import eu.trade.repo.search.codecs.impl.CMISStringPropertyTypeNormalizer;



public class JDBCDelegate {

//	private static final Logger LOG = LoggerFactory.getLogger(JDBCDelegate.class);
	
	private static final int PROPERTY_PAGE_SIZE = 5000;

	private static final String WRAP_PAGINATED_INIT_QUERY_PROPERTIES_PAGE = "select a.id, a.value from (select q.*, ROWNUM rn from ( ";
	private static final String WRAP_PAGINATED_END_QUERY_PROPERTIES_PAGE = " ) q where rownum < :end) a where rn >= :start";
	private static final String QUERY_PROPERTIES_PAGE = 
			WRAP_PAGINATED_INIT_QUERY_PROPERTIES_PAGE +	
			"select p.id, p.value from property p join object_type_property otp on p.object_type_property_id = otp.id where property_type='string' order by id" + 
			WRAP_PAGINATED_END_QUERY_PROPERTIES_PAGE;

	private static final String QUERY_UPDATE_PROPERTY = "update property set normalized_value = :normalizedValue where id = :id";

	
	
	private DataSourceTransactionManager transactionManager;
	private NamedParameterJdbcTemplate jdbcTemplate;
	
	
	public JDBCDelegate(DataSourceTransactionManager transactionManager, NamedParameterJdbcTemplate jdbcTemplate){
		this.transactionManager = transactionManager;
		this.jdbcTemplate = jdbcTemplate;
	}
	
	public void normalize(CMISStringPropertyTypeNormalizer normalizer){
		List<Property> propertyList = null;
		boolean done = false;
		
		long start = 0;
		while( !done ){
			propertyList = obtainPropertiesPage( start, PROPERTY_PAGE_SIZE);
			normalizePropertyList(propertyList, normalizer);
			done = ( propertyList.size() < PROPERTY_PAGE_SIZE -1);
			System.out.println(">>> normalized: start = " + start + "; pageSize = " + PROPERTY_PAGE_SIZE );
			start += PROPERTY_PAGE_SIZE;
		}
	}

	private List<Property> obtainPropertiesPage(long start, int pageSize){	
		MapSqlParameterSource namedParameters = new MapSqlParameterSource();
		namedParameters.addValue("start", start);
		namedParameters.addValue("end", start + pageSize);

		return jdbcTemplate.query(QUERY_PROPERTIES_PAGE, namedParameters, new ResultSetExtractor< List<Property> >() {

			@Override
			public List<Property> extractData(ResultSet rs) throws SQLException{

				List<Property> result = new ArrayList<Property>();
				while( rs.next() ){
					result.add(new Property(rs.getInt(1), rs.getString(2)));
				}
				return result;
			}
		});
	}


	public void normalizePropertyList(List<Property> propertyList, CMISStringPropertyTypeNormalizer normalizer){

		TransactionStatus txStatus = null;
		try {
			txStatus = transactionManager.getTransaction(new DefaultTransactionDefinition(TransactionDefinition.PROPAGATION_REQUIRES_NEW));

			List<Map<String, ?>> batch = new ArrayList<Map<String, ?>>();
			for (Property item : propertyList) {
				Map<String, Object> map = new HashMap<String, Object>();
				map.put("id", item.getId());
				map.put("normalizedValue", normalizer.normalizeString(item.getValue()));
				batch.add(map);
			}

			Map<String, Object>[] array = new HashMap[propertyList.size()];
			jdbcTemplate.batchUpdate(
					QUERY_UPDATE_PROPERTY,
					batch.toArray(array));

			commit(txStatus);
			System.out.println(">>> normalized properties number: " +  propertyList.size());
		} catch(Exception e) {
			e.printStackTrace();
			transactionManager.rollback(txStatus);
			throw e;
		}
	}

	//JDBC

	private void commit(TransactionStatus txStatus) {
		if (txStatus != null) {
			try{
				transactionManager.commit(txStatus);
			}catch(TransactionException te){
				te.printStackTrace();
				throw te;
			}
		}
	}


	private static class Property{
		private Integer id;
		private String value;
		private String normalizedValue;

		public Property(Integer id, String value){
			this.id = id;
			this.value = value;
		}

		public Integer getId() {
			return id;
		}
		public void setId(Integer id) {
			this.id = id;
		}
		public String getValue() {
			return value;
		}
		public void setValue(String value) {
			this.value = value;
		}
		public String getNormalizedValue() {
			return normalizedValue;
		}
		public void setNormalizedValue(String normalizedValue) {
			this.normalizedValue = normalizedValue;
		}


	}
}
